@@ -3,7 +3,7 @@
 ;; Copyright (C) 2006-2009, 2011-2012, 2015, 2016, 2017
 ;;   Phil Hagelberg, Doug Alcorn, Will Farrington, Chen Bin
 ;;
-;; Version: 5.5.2
+;; Version: 5.6.8
 ;; Author: Phil Hagelberg, Doug Alcorn, and Will Farrington
 ;; Maintainer: Chen Bin <chenbin.sh@gmail.com>
 ;; URL: https://github.com/technomancy/find-file-in-project
@@ -36,8 +36,10 @@
 ;; - Only dependency is GNU/BSD find
 ;; - Works on Windows with minimum setup
 ;; - Works flawlessly on Tramp Mode (https://www.emacswiki.org/emacs/TrampMode)
+;; - fd (faster alternative of find, see https://github.com/sharkdp/fd) is supported
 ;;
 ;; Usage,
+;;   - You can insert `(setq ffip-use-rust-fd t)' into ".emacs" to use fd (alternative of find)
 ;;   - `M-x find-file-in-project-at-point' guess the file name at point and
 ;;      find file
 ;;   - `M-x find-file-in-project-by-selected' use the selected region
@@ -62,6 +64,8 @@
 ;; .rb files that don't exist in the "vendor" directory.  In that case
 ;; you could set `ffip-find-options' to "-not -regex \".*vendor.*\"".
 
+;; `ffip-insert-file' insert file content into current buffer.
+
 ;; `find-file-with-similar-name' find file with similar name to current
 ;; opened file. The regular expression `ffip-strip-file-name-regex' is
 ;; also used by `find-file-with-similar-name'.
@@ -153,6 +157,14 @@
 (require 'diff-mode)
 (require 'windmove)
 
+(defvar ffip-use-rust-fd nil "Use use fd instead of find.")
+
+(defvar ffip-rust-fd-respect-ignore-files t
+  "Don 't show search results from '.*ignore' files")
+
+(defvar ffip-rust-fd-extra-opts ""
+  "rust fd extra options passed to cli.")
+
 (defvar ffip-window-ratio-alist
   '((1 . 1.61803398875)
     (2 . 2)
@@ -233,11 +245,16 @@ May be set using .dir-locals.el.  Checks each entry if set to a list.")
   "Use ido instead of ivy to display candidates.")
 
 (defvar ffip-patterns nil
-  "List of patterns to look for with `find-file-in-project'.")
+  "List of glob patterns to look for with `find-file-in-project'.")
 
 (defvar ffip-match-path-instead-of-filename nil
   "Match full path instead of file name when calling `find-file-in-project-by-selected'.")
 
+;; For "GNU/BSD Find", "*/test/*" matches "./test/" and "./dir/test/"
+;;
+;; But for "rust fd", only "test/*" matches "./test/" and "./dir/test/";
+;; "*/test/*" won't match "./test/" but matches "./dir/test/"
+;; Maybe it's bug of fd.
 (defvar ffip-prune-patterns
   '(;; VCS
     "*/.git/*"
@@ -304,13 +321,19 @@ May be set using .dir-locals.el.  Checks each entry if set to a list.")
     "*/.cask/*"
     ;; Python
     "*.pyc")
-  "List of directory/file patterns to not descend into when listing files in `find-file-in-project'.")
+  "List of directory/file glob patterns to not descend into when listing files in `find-file-in-project'.")
 
 (defvar ffip-find-options ""
   "Extra options to pass to `find' when using `find-file-in-project'.
 
 Use this to exclude portions of your project: \"-not -regex \\\".*svn.*\\\"\".")
 
+(defcustom ffip-find-pre-path-options ""
+  "Extra options to pass to `find' before path name options when using `find-file-in-project'.
+
+As required by `find', `-H', `-L', `-P', `-D' and `-O' must appear before the first path name, `.' for ffip's case.
+For example, use this to follow symbolic links inside your project: \"-L\".")
+
 (defvar ffip-project-root nil
   "If non-nil, overrides the project root directory location.")
 
@@ -323,7 +346,8 @@ This overrides variable `ffip-project-root' when set.")
   "Backup of `ivy-last'.  Requires ivy.")
 
 (defvar ffip-full-paths t
-  "If nil only file names in search results are visible.")
+  "If nil, search pattern only matches against file/directory name.
+If t,the pattern is matched against the full path.")
 
 (defvar ffip-debug nil "Print debug information.")
 
@@ -449,7 +473,7 @@ If CHECK-ONLY is true, only do the check."
      ((not keyword)
       (setq rlt ""))
      (t
-      (setq rlt (concat (if ffip-match-path-instead-of-filename "-iwholename" "-name")
+      (setq rlt (concat (if ffip-match-path-instead-of-filename "-iwholename" "-iname")
                         " \"*"
                         keyword
                         "*\"" ))))
@@ -475,9 +499,10 @@ If CHECK-ONLY is true, only do the check."
        (setq rlt exe))
      rlt))
 
-(defun ffip--executable-find (exe)
+(defun ffip--executable-find ()
   "Find EXE on all environments."
-  (let* (rlt)
+  (let* ((exe (if ffip-use-rust-fd "fd" "find"))
+         rlt)
     (cond
      ((file-remote-p default-directory)
       ;; In tramp mode and local windows, remote nix-like,
@@ -487,7 +512,17 @@ If CHECK-ONLY is true, only do the check."
       (setq rlt exe))
      ((setq rlt ffip-find-executable))
      ((eq system-type 'windows-nt)
-      (setq rlt (ffip--win-executable-find exe)))
+      ;; in case PATH is not setup properly
+      (cond
+       (ffip-use-rust-fd
+        (setq rlt (concat (getenv "USERPROFILE")
+                          "\\\\.cargo\\\\bin\\\\"
+                          exe
+                          ".exe"))
+        (unless (file-exists-p rlt)
+          (setq rlt exe)))
+       (t
+        (setq rlt (ffip--win-executable-find exe)))))
      ((setq rlt (executable-find exe)))
      (t
       ;; well, `executable-find' failed
@@ -496,15 +531,29 @@ If CHECK-ONLY is true, only do the check."
 
 (defun ffip--join-patterns (patterns)
   "Convert PATTERNS into cli arguments."
-  (if ffip-patterns
-      (format "\\( %s \\)" (mapconcat (lambda (pat) (format "-iwholename \"%s\"" pat))
-                         patterns " -or "))
-    ""))
+  (cond
+   ((and ffip-patterns (not ffip-use-rust-fd))
+    (format "\\( %s \\)" (mapconcat (lambda (pat) (format "-iwholename \"%s\"" pat))
+                                    patterns " -or ")))
+   (t
+    ;; rust fd only supports ONE pattern (and it's regular expression)
+    ;; which is precious resource to waste here
+    "")))
 
 (defun ffip--prune-patterns ()
   "Turn `ffip-prune-patterns' into a string that `find' can use."
-  (mapconcat (lambda (pat) (format "-iwholename \"%s\"" pat))
-             ffip-prune-patterns " -or "))
+  (cond
+   (ffip-use-rust-fd
+    ;; rust use "regular expression" which matches part of string
+    ;; while find use "glob pattern" which matches the whole string
+    ;; @see https://en.wikipedia.org/wiki/Glob_%28programming%29
+    (mapconcat (lambda (p)
+                 (format "-E \"%s\"" (replace-regexp-in-string "^\*/" "" p)))
+               ffip-prune-patterns " "))
+   (t
+    (mapconcat (lambda (p)
+                 (format "-iwholename \"%s\"" p))
+               ffip-prune-patterns " -or "))))
 
 ;;;###autoload
 (defun ffip-completing-read (prompt collection &optional action)
@@ -545,6 +594,52 @@ This function returns the selected candidate or nil."
     (ivy-read prompt collection
               :action action))))
 
+(defun ffip-create-shell-command (keyword is-finding-directory)
+  "Produce command to search KEYWORD.
+If IS-FINDING-DIRECTORY is t, we look up directory instead of file.
+Rust fd use regular expression.
+BSD/GNU Find use glob pattern."
+  (let* (cmd fmt tgt)
+    (cond
+     (ffip-use-rust-fd
+      ;; `-E` => exclude pattern
+      ;; `-c` => color
+      ;; `-i` => case insensitive
+      ;; `-t` => directory (d) or file (f)
+      ;; `-p` => match full path
+      (setq fmt (concat "%s %s -c never -i -t %s %s %s"
+                        (if ffip-full-paths " -p" "")
+                        (if ffip-rust-fd-respect-ignore-files "" " -I")
+                        " "
+                        ffip-rust-fd-extra-opts
+                        " %s"))
+      ;; fd use regular expression for target pattern (but glob pattern when excluding, sigh)
+      (setq tgt (if keyword (format "\".*%s\"" keyword) "")))
+     (t
+      (setq tgt
+            (if is-finding-directory (format "-iwholename \"*%s\"" keyword)
+              (ffip--create-filename-pattern-for-gnufind keyword)))
+      (setq fmt (concat "%s "
+                        ffip-find-pre-path-options
+                        " . \\( %s \\) -prune -o -type %s %s %s %s -print"))))
+
+    (setq cmd (format fmt
+                      (ffip--executable-find)
+                      (ffip--prune-patterns)
+                      (if is-finding-directory "d" "f")
+                      ;; When finding directory, the keyword is like:
+                      ;; "proj/hello/world"
+                      ffip-find-options
+                      (ffip--join-patterns ffip-patterns)
+                      tgt))
+    cmd))
+
+(defun ffip-glob-to-regex (s)
+  "Convert glob pattern S into regular expression."
+  (setq s (replace-regexp-in-string "\\." "\\\\." s))
+  (setq s (replace-regexp-in-string "\*" ".*" s))
+  s)
+
 ;;;###autoload
 (defun ffip-project-search (keyword is-finding-directory &optional directory-to-search)
   "Return an alist of all filenames in the project and their path.
@@ -560,18 +655,17 @@ DIRECTORY-TO-SEARCH specify the root directory to search."
          (root (or directory-to-search
                    (ffip-get-project-root-directory)))
          (default-directory (file-name-as-directory root))
-         (cmd (format "%s . \\( %s \\) -prune -o -type %s %s %s %s -print"
-                      (ffip--executable-find "find")
-                      (ffip--prune-patterns)
-                      (if is-finding-directory "d" "f")
-                      (ffip--join-patterns ffip-patterns)
-                      ;; When finding directory, the keyword is like:
-                      ;; "proj/hello/world"
-                      (if is-finding-directory (format "-iwholename \"*%s\"" keyword)
-                        (ffip--create-filename-pattern-for-gnufind keyword))
-                      ffip-find-options)))
+         (cmd (ffip-create-shell-command keyword is-finding-directory))
+         (collection (split-string (ffip-shell-command-to-string cmd) "[\r\n]+" t)))
 
-    (if ffip-debug (message "run cmd at %s: %s" default-directory cmd))
+    (if ffip-debug (message "run command at %s: %s" default-directory cmd))
+
+    (when (and ffip-use-rust-fd ffip-patterns)
+      ;; filter result with Lisp because fd does NOT support multiple patterns
+      (let* ((r (concat "^" (mapconcat 'ffip-glob-to-regex ffip-patterns "\\|") "$")))
+        (setq collection (delq nil (mapcar (lambda (s)
+                                             (if (string-match-p r s) s))
+                                           collection)))))
     (setq rlt
           (mapcar (lambda (file)
                     (if ffip-full-paths
@@ -580,7 +674,7 @@ DIRECTORY-TO-SEARCH specify the root directory to search."
                       (cons (file-name-nondirectory file)
                             (expand-file-name file))))
                   ;; #15 improving handling of directories containing space
-                  (split-string (ffip-shell-command-to-string cmd) "[\r\n]+" t)))
+                  collection))
     rlt))
 
 (defun ffip--forward-line (lnum)
@@ -640,8 +734,10 @@ This function is the API to find files."
 
 (defun ffip-read-keyword ()
   "Read keyword from selected text or user input."
-  (if (region-active-p) (ffip--read-selected)
-    (read-string "Enter keyword (or press ENTER):")))
+  (let* ((hint (if ffip-use-rust-fd "Enter regex (or press ENTER):"
+                 "Enter keyword (or press ENTER):")))
+    (if (region-active-p) (ffip--read-selected)
+      (read-string hint))))
 
 ;;;###autoload
 (defun ffip-create-project-file ()
@@ -776,6 +872,23 @@ If OPEN-ANOTHER-WINDOW is not nil, the file will be opened in new window."
   (interactive "P")
   (ffip-find-files (ffip-read-keyword) open-another-window))
 
+;;;###autoload
+(defun ffip-insert-file ()
+  "Insert contents of file in current buffer.
+The file name is selected interactively from candidates in project."
+  (interactive)
+  (let* ((cands (ffip-project-search (ffip-read-keyword) nil))
+         root)
+    (when (> (length cands) 0)
+      (setq root (file-name-nondirectory (directory-file-name (ffip-get-project-root-directory))))
+      (ffip-completing-read
+       (format "Read file in %s/: " root)
+       cands
+       `(lambda (file)
+          ;; only one item in project files
+          (if (listp file) (setq file (cdr file)))
+          (insert-file file))))))
+
 ;;;###autoload
 (defun find-file-with-similar-name (&optional open-another-window)
   "Use base name of current file as keyword which could be further stripped.
@@ -1044,6 +1157,23 @@ If NUM is not nil, the corresponding backend is executed directly."
 
 (defalias 'ffip-show-diff 'ffip-show-diff-by-description)
 
+(defadvice read-file-name (around ffip-read-file-name-hack activate)
+  (cond
+   (ffip-read-file-name-hijacked-p
+    ;; only hack read-file-name once
+    (setq ffip-read-file-name-hijacked-p nil)
+    (let* ((args (ad-get-args 0))
+           (file-name (file-name-nondirectory (nth 2 args)))
+           (default-directory (ffip-project-root))
+           (cands (ffip-project-search file-name nil default-directory))
+           (rlt (if cands (ffip-completing-read "Files: " cands))))
+      (when rlt
+        (setq rlt (file-truename rlt))
+        (run-hook-with-args 'ffip-diff-apply-hunk-hook rlt)
+        (setq ad-return-value rlt))))
+   (t
+    ad-do-it)))
+
 ;;;###autoload
 (defun ffip-diff-apply-hunk (&optional reverse)
   "Apply current hunk in `diff-mode'.  Try to locate the file to patch.
@@ -1051,23 +1181,13 @@ Similar to `diff-apply-hunk' but smarter.
 Please read documenation of `diff-apply-hunk' to get more details.
 If REVERSE is t, applied patch is reverted."
   (interactive "P")
-  (setq ffip-read-file-name-hijacked-p t)
-  (defadvice read-file-name (around ffip-read-file-name-hack activate)
-    (cond
-     (ffip-read-file-name-hijacked-p
-      (let* ((args (ad-get-args 0))
-             (file-name (file-name-nondirectory (nth 2 args)))
-             (default-directory (ffip-project-root))
-             (cands (ffip-project-search file-name nil default-directory))
-             (rlt (if cands (ffip-completing-read "Files: " cands))))
-        (when rlt
-          (setq rlt (file-truename rlt))
-          (run-hook-with-args 'ffip-diff-apply-hunk-hook rlt)
-          (setq ad-return-value rlt))))
-     (t
-      ad-do-it)))
-  (diff-apply-hunk reverse)
-  (setq ffip-read-file-name-hijacked-p nil))
+  (cond
+   ((derived-mode-p 'diff-mode)
+    (setq ffip-read-file-name-hijacked-p t)
+    (diff-apply-hunk reverse)
+    (setq ffip-read-file-name-hijacked-p nil))
+   (t
+    (message "This command only run in `diff-mode' and `ffip-diff-mode'."))))
 
 ;; safe locals
 (progn
